/*
 * s_hanoi.c
 *
 *	2016/06/03
 *
 */

#ifndef __LINUX__
#include <windows.h>
#endif

/*
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <GL/freeglut.h>

/*
 *
 */

#include "s_const.h"
#include "s_alloc.h"
#include "s_que.h"
#include "s_thread_manager.h"
#include "s_hanoi.h"

/*
 *
 */

#define	S_POLE_WIDTH		20
#define	S_MAX_DISC_SIZE		100
#define	S_POLE_HIGHT		(S_MAX_DISC_SIZE * 3 + S_POLE_WIDTH)

#define	S_BASE_HIGHT		10
#define	S_BASE_WIDTH		( ( S_MAX_DISC_SIZE + S_POLE_WIDTH ) * S_NUMBER_OF_POLES + S_POLE_WIDTH )

#define S_HANOI_SCALE		( ( S_BASE_WIDTH + S_POLE_WIDTH * 2 ) / 2 )
#define S_HANOI_CENTER		( S_HANOI_SCALE / 2 )

/*
 *
 */

#define	S_NUMBER_OF_POLES	3
#define	S_INIT_DISCS		3

static int S_discs = S_INIT_DISCS;
static int *S_tower_body[S_NUMBER_OF_POLES] = { NULL, NULL, NULL };
static int S_tower_top[S_NUMBER_OF_POLES] = { 0, 0, 0 };
static int S_discs_hight;
static int S_discs_width_step;

/*
 *
 */

static int S_hanoi_display_flag = NO;
static int S_hanoi_display_animation = NO;

#define	S_MOVE_STEP		1

#define S_MOVE_STAY		0
#define S_MOVE_UP		1
#define S_MOVE_LEFT		2
#define S_MOVE_RIGHT	3
#define S_MOVE_DOWN		4

static int S_move_mode = S_MOVE_STAY;

static int S_move_size;
static int S_move_from;
static int S_move_to;

static int S_move_x;
static int S_move_y;

/*
 *
 */

static void S_hanoi_clear( int discs ) {
  int i;

  S_hanoi_display_animation = NO;
  S_move_mode = S_MOVE_STAY;

  for ( i = 0; i < S_NUMBER_OF_POLES; i++ ) {
	S_tower_body[i][0] = 0;
	S_tower_top[i] = 0;
  }

  for ( i = 0; i < discs; i++ ) {
	S_tower_body[0][i] = discs - i;
	S_tower_top[0] = discs;
  }

}

/*
 *
 */

static void S_hanoi_unsetup( void ) {
  int i;

  if ( S_tower_body[0] == NULL ) {
	for ( i = 0; i < S_NUMBER_OF_POLES; i++ ) {
	  s_free ( S_tower_body[i] );
	}
	S_tower_body[0] = NULL;
  }
}

static void S_hanoi_setup( int discs ) {
  int i;

  if ( S_tower_body[0] == NULL ) {
	for ( i = 0; i < S_NUMBER_OF_POLES; i++ ) {
	  S_tower_body[i] = s_calloc ( discs, int );
	}
  }

  S_discs_hight = (S_MAX_DISC_SIZE * 3) / discs;
  S_discs_width_step = (S_MAX_DISC_SIZE-S_POLE_WIDTH) / (discs-1);

  S_hanoi_clear( discs );
}

/*
 *
 */

static void S_hanoi_rect ( int px, int py, int sx, int sy ) {

  px -= S_HANOI_CENTER;
  py -= S_HANOI_SCALE - S_POLE_WIDTH;

  glBegin(GL_QUADS);                   // 開始 四角形
  glVertex2d((double)px/S_HANOI_SCALE, (double)py/S_HANOI_SCALE);
  glVertex2d((double)(px+sx)/S_HANOI_SCALE, (double)py/S_HANOI_SCALE);
  glVertex2d((double)(px+sx)/S_HANOI_SCALE, (double)(py+sy)/S_HANOI_SCALE);
  glVertex2d((double)(px)/S_HANOI_SCALE, (double)(py+sy)/S_HANOI_SCALE);
  glEnd();

}

static void S_hanoi_crect ( int cx, int cy, int sx, int sy ) {

  S_hanoi_rect ( cx - sx / 2, cy, sx, sy );
}

static int S_pole_pos ( int pos ) {

  return S_HANOI_CENTER + ( pos - 1 ) * ( S_MAX_DISC_SIZE + S_POLE_WIDTH );
}

static int S_desc_width ( int size ) {

  return S_POLE_WIDTH * 3 + S_discs_width_step * ( size - 1 );
}

static int S_desc_hight ( int i ) {

  return S_BASE_HIGHT + i * ( S_discs_hight + 2 );
}

static void S_hanoi_pole ( int pos ) {
  int cx = S_pole_pos ( pos );

  S_hanoi_crect ( cx, S_BASE_HIGHT, S_POLE_WIDTH, S_POLE_HIGHT );
}

static void S_pole_display(void) {
  int i;

  glColor3d(0.0, 1.0, 1.0);
  S_hanoi_crect ( S_HANOI_CENTER, 0, S_BASE_WIDTH, S_BASE_HIGHT );

  for ( i = 0; i < S_NUMBER_OF_POLES; i++ ) {
	S_hanoi_pole ( i );
  }
}

static void S_moving_desc_display() {

  if ( S_move_mode != S_MOVE_STAY ) {
	S_hanoi_crect ( S_move_x, S_move_y, S_desc_width ( S_move_size ), S_discs_hight );
  }

}


static void S_pole_desc_display( int pos ) {
  int cx = S_pole_pos ( pos );
  int i;

  for ( i = 0; i < S_tower_top[pos]; i++ ) {
	S_hanoi_crect ( cx, S_desc_hight( i ), S_desc_width ( S_tower_body[pos][i] ), S_discs_hight );
  }
}

static void S_desc_display(void) {
  int i;

  glColor3d(0.5, 0.5, 1.0);

  for ( i = 0; i < S_NUMBER_OF_POLES; i++ ) {
	S_pole_desc_display( i );
  }
}

static void S_hanoi_display() {

  if ( S_hanoi_display_flag == YES ) {
	usleep ( 1000 );
	return;
  } else {
	S_hanoi_display_flag = YES;
  }

  glClear(GL_COLOR_BUFFER_BIT);

  S_pole_display();
  S_desc_display();
  S_moving_desc_display();

  glFlush();

  S_hanoi_display_flag = NO;

}

void S_hanoi_graphics_init(void)
{
  glClearColor(0.0, 0.0, 1.0, 1.0);
}

/*
 *
 */

static S_Que *que = NULL;

static void S_hanoi_deque( void ) {

  if ( que -> finish == NO ) {
	//	fprintf ( stderr, "Dequle\n" );

	if ( S_hanoi_display_animation == YES ) {
	  usleep ( 1000 );
	  switch ( S_move_mode ) {
	  case S_MOVE_UP:
		if ( S_move_y < S_POLE_HIGHT + S_POLE_WIDTH ) {
		  S_move_y += S_MOVE_STEP;
		} else if ( S_move_to == S_move_from ) {
		  S_hanoi_display_animation = NO;
		} else if ( S_move_to < S_move_from ) {
		  S_move_mode = S_MOVE_LEFT;
		} else {
		  S_move_mode = S_MOVE_RIGHT;
		}
		break;
	  case S_MOVE_DOWN:
		S_move_y -= S_MOVE_STEP;
		if ( S_move_y < S_desc_hight ( S_tower_top[S_move_to] ) ) {
		  S_tower_body[S_move_to][S_tower_top[S_move_to]++] = S_move_size;
		  S_hanoi_display_animation = NO;
		  S_move_mode = S_MOVE_STAY;
		}
		break;
	  case S_MOVE_LEFT:
		S_move_x -= S_MOVE_STEP;
		if ( S_move_x <= S_pole_pos ( S_move_to ) ) {
		  S_move_x = S_pole_pos ( S_move_to );
		  S_move_mode = S_MOVE_DOWN;
		}
		break;
	  case S_MOVE_RIGHT:
		S_move_x += S_MOVE_STEP;
		if ( S_move_x >= S_pole_pos ( S_move_to ) ) {
		  S_move_x = S_pole_pos ( S_move_to );
		  S_move_mode = S_MOVE_DOWN;
		}
		break;
	  }
	  glutPostRedisplay();
	} else if ( !s_is_empty_que ( que ) ) {
	  	int data = s_deque ( que );
		int f;
		int t;
		int d;

		switch ( data ) {
		case S_HANOI_CLEAR:
		  S_hanoi_clear(S_discs);
		  break;
		case S_HANOI_MOVE:
		  f = s_deque( que );
		  t = s_deque( que );

		  if ( f >= S_discs || f < 0 || t >= S_discs || t < 0 ) {
			break;
		  }

		  if ( f != t ) {
			if ( S_tower_top [ f ] > 0 ) {
			  d = S_tower_body[f] [ S_tower_top [ f ] - 1 ];
			  if ( S_tower_top [ t ] == 0 || S_tower_body[t][S_tower_top[ t ]-1] > d ) {
				S_move_size = d;
				S_tower_top [ f ]--;
				S_move_from = f;
				S_move_to = t;
				S_move_mode = S_MOVE_UP;
				S_move_x = S_pole_pos ( f );
				S_move_y = S_desc_hight(S_tower_top[f]);
				S_hanoi_display_animation = YES;
			  }
			}
		  }
		  break;
		case S_HANOI_SIZE:
			data = s_deque( que );
			if ( data > 1 ) {
			  S_hanoi_unsetup();
			  S_discs = data;
			  S_hanoi_setup( data );
			}
		  break;
		}

		glutPostRedisplay();
	}
  }
}

/*
 *
 */

static S_hanoi_init_flag = 0;

static void S_hanoi_exit(void) {

  //  fprintf ( stderr, "call S_hanoi_exit(%d)\n", S_hanoi_init_flag );

  if ( S_hanoi_init_flag == 1 ) {
	glutLeaveMainLoop();	// stop main loop
	if ( S_hanoi_init_flag != 0 ) {
	  //  fprintf ( stderr, "S_hanoi_exit waiting..\n" );
	  usleep ( 1000 );
	}
	S_hanoi_unsetup();
  }
}

/*
 *
 */

// キーイベント関数
static void s_hanoi_keyboard(unsigned char key, int x, int y)
{
  int pos;

	switch (key) {
		case 'q':
		case 'Q':
		case 0x1b:
		  S_hanoi_unsetup();
		  exit(0);	// （このあと終了処理関数が呼ばれる）
		case 'c':
		case 'C':
  		  S_hanoi_clear(S_discs);
		  break;
		case '1':
		case '2':
		case '3':
		  pos = key - '1';
		  switch ( S_move_mode ) {
		  case S_MOVE_STAY:
			if ( S_tower_top[pos] > 0 ) {
			  S_move_from = pos;
			  S_move_to = pos;
			  S_move_size = S_tower_body[pos][--S_tower_top[pos]];
			  S_move_x = S_pole_pos ( pos );
			  S_move_y = S_desc_hight(S_tower_top[pos]);
			  S_move_mode = S_MOVE_UP;
			  S_hanoi_display_animation = YES;
			}
			break;
		  case S_MOVE_UP:
			if ( S_move_from == pos ) {
			  S_move_mode = S_MOVE_DOWN;
			  S_hanoi_display_animation = YES;
			  S_move_to = pos;
			} else if ( S_tower_top[pos] == 0 || S_move_size < S_tower_body[pos][S_tower_top[pos]-1] ) {
			  S_move_to = pos;
			  if ( S_move_to < S_move_from ) {
				S_move_mode = S_MOVE_LEFT;
			  } else {
				S_move_mode = S_MOVE_RIGHT;
			  }
			  S_hanoi_display_animation = YES;
			}
			break;
		  case S_MOVE_LEFT:
			  break;
		  case S_MOVE_RIGHT:
			  break;
		  case S_MOVE_DOWN:
			  break;
		  }
		  break;
		default:
			break;
	}
}

/*
 *
 */

static void *S_hanoi_thread ( void *pParam ) {
  char *argv[] = { "Hanoi", NULL };
  int argc = 1;

  //  fprintf ( stderr, "Start: S_hanoi_thread\n" );

  S_hanoi_setup( S_INIT_DISCS );

  que = (S_Que *)pParam;

  glutInit(&argc, argv);
  glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE,GLUT_ACTION_GLUTMAINLOOP_RETURNS);
  glutInitDisplayMode(GLUT_RGBA);
  glutCreateWindow(argv[0]);
  glutDisplayFunc(S_hanoi_display);
  glutKeyboardFunc(s_hanoi_keyboard);	// キー入力時の処理関数の設定
  glutIdleFunc(S_hanoi_deque);
  S_hanoi_graphics_init();
  glutMainLoop();

  //  fprintf ( stderr, "End: S_hanoi_thread\n" );

  S_hanoi_init_flag = 0;

  return pParam;
}

void s_hanoi_init(void) {

  if ( S_hanoi_init_flag == 0 ) {
	S_hanoi_init_flag = 1;
	que = s_thread_manager_add ( S_hanoi_thread, 100, S_hanoi_exit );
  }
}

/*
 *
 */

void s_hanoi_size ( char *discs ) {
	s_hanoi_init();
	s_enque ( que, S_HANOI_SIZE );
	s_enque ( que, strlen ( discs ) );
}

void s_hanoi_clear(void) {
	s_hanoi_init();
	s_enque ( que, S_HANOI_CLEAR );
}

void s_hanoi_move ( char *from, char *to ) {
	s_hanoi_init();
	s_enque ( que, S_HANOI_MOVE );
	s_enque ( que, *from - '1' );
	s_enque ( que, *to - '1' );
}

/*
 *
 */

void s_hanoi_stop(void) {

  if ( S_hanoi_init_flag == 1 ) {
	glutLeaveMainLoop();	// stop main loop
  }
  que -> finish = YES;	// stop que access

}

/*
 *
 */

